import io
import os


class Service:
    __bucket_class__ = None

    requires_location = True

    def __init__(self, id, default_location=None):
        self.id = id
        self.default_location = default_location
        print(f"Service: Set default location {default_location}")

    def __str__(self):
        return "{}://".format(self.id)

    def bucket(self, name, location=None):
        if self.requires_location and (location is None and self.default_location is None):
            raise Exception("No location or default location set!")

        if self.__bucket_class__ is None:
            raise Exception("No __bucket_class__ was set for {}".format(self))

        return self.__bucket_class__(self, name, location or self.default_location)

    def __eq__(self, other):
        if not isinstance(other, Service):
            return False

        return self.id == other.id


class Bucket:
    def __init__(self, service: Service, name: str, location: str):
        self.service = service
        self.name = name
        self.location = location

    def __repr__(self):
        return f'<{self.__class__.__name__} {str(self)}>'

    def __str__(self):
        return "{}{}/{}".format(self.service, self.location, self.name)

    def cubby(self, name):
        raise NotImplementedError()

    def delete(self):
        raise NotImplementedError()

    def create(self):
        raise NotImplementedError()

    def list(self, prefix=None, max_keys=None):
        raise NotImplementedError()


class Cubby:
    def __init__(self, bucket, key):
        self.bucket = bucket
        self.key = key

    def __repr__(self):
        return f'<{self.__class__.__name__} {str(self)}>'

    def __str__(self):
        return "{}/{}".format(self.bucket, self.key)

    def retrieve(self, filepath=None, file=None):
        if filepath is not None:
            if os.path.isdir(filepath):
                filepath = os.path.join(filepath, self.key)

            with open(filepath, 'wb') as file:
                return self.retrieve_filelike(file)
        elif file is not None:
            return self.retrieve_filelike(file)
        else:
            buffer = io.BytesIO()
            self.retrieve_filelike(buffer)
            buffer.seek(0)
            return buffer.read()

    def retrieve_filelike(self, file):
        raise NotImplementedError()

    def store(self, filepath=None, file=None, string=None, bytes=None, encoding='utf-8'):
        if filepath is not None:
            with open(filepath, 'rb') as file:
                return self.store_filelike(file)
        elif file is not None:
            self.store_filelike(file)
        elif string is not None:
            self.store_filelike(io.BytesIO(string.encode(encoding)))
        elif bytes is not None:
            self.store_filelike(io.BytesIO(bytes))
        else:
            raise Exception("One of [filepath, file, string, bytes] must be specified.")

        return self

    def store_filelike(self, filelike):
        raise NotImplementedError()

    DefaultUrlExpiration = None

    def url(self, expiration=DefaultUrlExpiration):
        raise NotImplementedError()

    def filesize(self):
        raise NotImplementedError()

    def delete(self):
        raise NotImplementedError()

    def exists(self):
        raise NotImplementedError()

    def copy_to_native_cubby(self, cubby=None):
        """Copies from a Cubby of this type to another cubby of this type."""
        raise NotImplementedError()

    def copy_to(self, key=None, cubby=None):
        if key:
            cubby = self.bucket.cubby(key)

        if isinstance(cubby, self.__class__):
            self.copy_to_native_cubby(cubby)
            return cubby

        cubby.store_filelike(self.retrieve_filelike())
        return cubby

    def move_to(self, key=None, cubby=None):
        copy = self.copy_to(key=key, cubby=cubby)
        self.delete()
        return copy

    @property
    def extension(self):
        return os.path.splitext(self.key)[1]

    @property
    def service(self):
        return self.bucket.service



import re

from flask import current_app

from .backends import Service, FileService, S3Service


# A regex for bucket strings like s3://us-west-1/bucket
WAREHOUSE_BUCKET_REGEX = \
    re.compile(r"(?P<service>(s3|file)):\/\/(?P<location>[^\/]+)?\/(?P<bucket>[^\/]+)")


# A regex for cubby strings like s3://us-west-1/bucket/key
WAREHOUSE_CUBBY_REGEX = \
    re.compile(r"{}\/(?P<key>.+)".format(WAREHOUSE_BUCKET_REGEX.pattern))


class Warehouse(Service):
    """
    Clean abstraction over several file storage backends (S3, Alicloud, local).
    """

    def __init__(self, app=None):
        self.services = {
            's3': S3Service,
            'file': FileService,
        }

        self.service: Service = None
        self.default_bucket = None

        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        self.app = app

        app.config.setdefault('WAREHOUSE_DEFAULT_SERVICE', 'file')

        default_service_key = app.config['WAREHOUSE_DEFAULT_SERVICE']

        self.default_location = app.config.get('WAREHOUSE_DEFAULT_LOCATION')
        self.default_bucket = app.config.get('WAREHOUSE_DEFAULT_BUCKET')

        self.service = self._create_service(service=default_service_key,
                                            location=self.default_location,
                                            app=app)

    def bucket(self, name=None, location=None):
        if self.app is None:
            raise RuntimeError("Storage.init_app() was not called!")

        if self.default_bucket is None and name is None:
            raise Exception("'WAREHOUSE_DEFAULT_BUCKET' is not set!")

        return self.service.bucket(name or self.default_bucket, location or self.default_location)

    def _create_service(self, service=None, location=None, app=None):
        try:
            service_constructor = self.services[service]
        except KeyError:
            raise Exception("No StorageService was registered named '{}'".format(service))

        return service_constructor(app or current_app, default_location=location or self.default_location)

    def _create_bucket_or_cubby(self, service=None, location=None, bucket=None, key=None, app=None):
        service = self._create_service(service=service, location=location)
        bucket = service.bucket(bucket)

        if key is None:
            return bucket

        return bucket.cubby(key)

    def __call__(self, bucket_or_key_str):
        match = (WAREHOUSE_CUBBY_REGEX.match(bucket_or_key_str)
                 or WAREHOUSE_BUCKET_REGEX.match(bucket_or_key_str))

        if match:
            return self._create_bucket_or_cubby(**match.groupdict())

        raise Exception("Could not parse '{}' as a Bucket or Cubby str".format(bucket_or_key_str))

    def __repr__(self):
        return "<Warehouse service={} default_bucket={}>".format(self.service, self.default_bucket)

    def __str__(self):
        return repr(self)


__all__ = ["Warehouse"]



# -*- coding: utf-8 -*-

"""
    flaskext.flask-warehouse
    ---------------

    A clean abstraction over cloud file storage platforms like S3, Alicloud, or
    Heroku.

    :copyright: (c) by Joshua Augustin.
    :license: MIT license , see LICENSE for more details.
"""


__author__ = """Joshua Augustin"""
__email__ = 'augustinspring@gmail.com'
__version__ = '0.1.8'


from .flask_warehouse import Warehouse
assert(Warehouse)



import os
import shutil

from flask import Flask, url_for

from .service import Bucket, Cubby, Service


class FileService(Service):
    requires_location = False

    def __init__(self, app: Flask, default_location=None):
        super().__init__('file', default_location=default_location)

        self.root = app.static_folder

        self.abspath = os.path.abspath(self.root)

        if not os.path.isdir(self.abspath):
            os.makedirs(self.abspath)


class FolderBucket(Bucket):
    def __init__(self, service: FileService, name: str, location: str):
        super().__init__(service, name, location)

        self.abspath = os.path.abspath(os.path.join(service.root, name))

        if not os.path.isdir(self.abspath):
            os.makedirs(self.abspath)

    def cubby(self, name, content_type=None, acl='public-read'):
        return FileCubby(self, name, content_type=content_type, acl=acl)

    def delete(self):
        shutil.rmtree(self.abspath)

    def list(self, prefix=None, max_keys=None, **kwargs):
        return [FileCubby(self, key) for key in os.listdir(self.abspath)]


class FileCubby(Cubby):
    def __init__(self, bucket: FolderBucket, name: str, content_type=None, acl='public-read'):
        super().__init__(bucket, name)

        if not os.path.isdir(self.dirpath()):
            os.makedirs(self.dirpath())

        self.content_type = content_type
        self.acl = acl

    def dirpath(self):
        return os.path.dirname(self.filepath())

    def filepath(self):
        return os.path.join(self.bucket.abspath, self.key)

    def keypath(self):
        return os.path.join(self.bucket.name, self.key)

    def url(self, duration=Cubby.DefaultUrlExpiration):
        return url_for('static', filename=self.keypath(), _external=True)

    def store_filelike(self, filelike):
        shutil.copyfileobj(filelike, open(self.filepath(), 'wb'))

    def retrieve_filelike(self, filelike):
        shutil.copyfileobj(open(self.filepath(), 'rb'), filelike)

    def delete(self):
        if self.exists():
            os.remove(self.filepath())

        return not self.exists()

    def filesize(self):
        return os.path.getsize(self.filepath())

    def service_id(self):
        return "file"

    def exists(self):
        return os.path.isfile(self.filepath())

    def copy_to_native_cubby(self, cubby=None):
        shutil.copy(self.filepath(), cubby.filepath())


FileService.__bucket_class__ = FolderBucket



import gzip

from tempfile import SpooledTemporaryFile

from flask import Flask

import boto3

from .service import Bucket, Cubby, Service


class S3Service(Service):
    def __init__(self, app: Flask, aws_access_key_id=None, aws_secret_access_key=None, default_location=None):
        super().__init__('s3', default_location=default_location)

        self.session = boto3.Session(aws_access_key_id=aws_access_key_id,
                                     aws_secret_access_key=aws_secret_access_key,
                                     region_name=default_location)

        self.s3 = boto3.resource('s3')
        self.client = boto3.client('s3')

        try:
            self.client.list_buckets()
        except Exception:
            raise Exception("Failed to connect to S3 - ensure that AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY are both set.")


class S3Bucket(Bucket):
    def __init__(self, service: S3Service, name: str, location: str):
        super().__init__(service, name, location)

        self._bucket: boto3.s3.bucket.Bucket = self.service.s3.Bucket(name)

    def cubby(self, name, content_type=None, acl='public-read'):
        return S3Cubby(self, name, content_type=content_type, acl=acl)

    def delete(self):
        self._bucket.delete()

    def create(self):
        bucket_configuration = {'LocationConstraint': self.location}
        self._bucket.create(CreateBucketConfiguration=bucket_configuration)

    def list(self, prefix=None, max_keys=None, **kwargs):
        if prefix is not None:
            kwargs['Prefix'] = prefix
        if max_keys is not None:
            kwargs['MaxKeys'] = max_keys

        keys = self._bucket.objects.filter(**kwargs)

        return [S3Cubby(self, key.key) for key in keys]

    def __eq__(self, other):
        if not isinstance(other, S3Bucket):
            return False
        
        return (self.service == other.service 
        and self.name == other.name 
        and self.location == other.location)

    def copy_key(self, dst_key: str, src_key: str, src_bucket_name: str = None):
        if src_bucket_name is None:
            src_bucket_name = self.name

        self._bucket.copy({"Bucket":src_bucket_name, "Key": src_key}, dst_key)


class S3Cubby(Cubby):
    def __init__(self, bucket: S3Bucket, name, content_type=None, acl=None, key=None):
        super().__init__(bucket, name)

        self.content_type = content_type

        if key is None:
            self._key = self.bucket._bucket.Object(name)
        else:
            self._key = key
            self.key = self._key.nameG

        self.acl = acl

    @staticmethod
    def apply_func_filelike(filelike, fn):
        """Applies a function to content in a file like object.

            fn: a function that accepts a single argument of the original content, and returns a single
                variable as the updated content.
        """
        content = filelike.read()
        content = fn(content)

        # replace the content in the file like object
        filelike.seek(0)
        filelike.write(content)
        filelike.truncate()

        # seek back to the start so the filelike object is usable
        filelike.seek(0)

    def store_filelike(self, filelike, tempcopy=False):
        if tempcopy:
            copy = SpooledTemporaryFile()  # boto3 now closes the file.
            copy.write(filelike.read())
            copy.seek(0)

            filelike = copy

        ExtraArgs = {}

        if self.acl:
            ExtraArgs['ACL'] = self.acl

        new_content_type = self.content_type
        existing_content_type = (self.exists() and self.mimetype()) or None
        if new_content_type or existing_content_type:
            ExtraArgs['ContentType'] = new_content_type or existing_content_type

        if self.exists() and self.content_encoding() == "gzip":
            self.apply_func_filelike(filelike, fn=gzip.compress)
            ExtraArgs['ContentEncoding'] = self.content_encoding(reload=False)

        self._key.upload_fileobj(filelike, ExtraArgs=ExtraArgs)

        return self.url()

    def retrieve_filelike(self, filelike):
        if filelike.closed:
            raise Exception("File provided was already closed.")

        self._key.download_fileobj(filelike)

        # when ContentEncoding is set, 'download_fileobj' seems to read the filelike
        # object, so that's why the next line is needed, there's an open issue on moto for that
        # ref: https://github.com/spulec/moto/issues/2926
        filelike.seek(0)

        if self.content_encoding() == "gzip":
            self.apply_func_filelike(filelike, fn=gzip.decompress)

        return

    def url(self, expiration=Cubby.DefaultUrlExpiration):
        service: S3Service = self.bucket.service
        result = service.client.generate_presigned_url('get_object',
                                                       Params={
                                                           "Bucket": self.bucket.name,
                                                           "Key": self.key
                                                       },
                                                       ExpiresIn=int(expiration.total_seconds()) if expiration else None)

        if expiration is None:
            # chop off ?AWSAccessKeyId= etc. - if public, this will work.
            result = result.split('?')[0]

        return result

    def delete(self):
        self._key.delete()
        return not self.exists()

    def filesize(self, reload=True):
        if reload:
            self._key.reload()

        return self._key.content_length

    def exists(self):
        matches = list(self.bucket._bucket.objects.filter(Prefix=self.key))
        return len(matches) > 0 and matches[0].key == self.key

    def metadata(self, reload=True):
        if reload:
            self._key.reload()

        return self._key.metadata

    def mimetype(self, reload=True):
        if reload:
            self._key.reload()

        return self._key.content_type

    def set_mimetype(self, mimetype):
        if not self.acl:
            raise Exception(
                "S3Cubby's acl must be set or it will be removed by set_mimetype()!"
            )

        args = dict(
            CopySource={"Bucket": self.bucket.name, "Key": self.key},
            MetadataDirective="REPLACE",
            ACL=self.acl,
            ContentType=mimetype,
        )

        if self.content_encoding():
            args["ContentEncoding"] = self.content_encoding(reload=False)

        self._key.copy_from(**args)

        return self.mimetype()

    def content_encoding(self, reload=True):
        if reload:
            self._key.reload()

        return self._key.content_encoding

    def set_content_encoding(self, content_encoding):
        if not self.acl:
            raise Exception(
                "S3Cubby's acl must be set or it will be removed by set_content_encoding()!"
            )

        args = dict(
            CopySource={"Bucket": self.bucket.name, "Key": self.key},
            MetadataDirective="REPLACE",
            ACL=self.acl,
            ContentEncoding=content_encoding,
        )

        if self.mimetype():
            args["ContentType"] = self.mimetype(reload=False)

        self._key.copy_from(**args)
        return self.content_encoding()

    def set_metadata(self, metadata: dict = {}):
        self._key.copy_from(CopySource={'Bucket': self.bucket.name, 'Key': self.key},
                            MetadataDirective="REPLACE",
                            Metadata=metadata)
        self._key.reload()
        return metadata

    def __eq__(self, other):
        if not isinstance(other, S3Cubby):
            return False
        
        return self.bucket == other.bucket and self.key == other.key

    def copy_to_native_cubby(self, cubby=None):
        cubby.bucket.copy_key(cubby.key, self.key, src_bucket_name=self.bucket.name)


S3Service.__bucket_class__ = S3Bucket



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'Flask>=1.0',
]

test_requirements = [
    'Flask>=1.0',
]

setup(
    name='flask_warehouse',
    version='0.1.8',
    description="A clean abstraction over cloud file storage platforms like S3, Alicloud, or Heroku.",
    long_description=readme + '\n\n' + history,
    author="Joshua Augustin",
    author_email='augustinspring@gmail.com',
    url='https://github.com/augustjd/flask-warehouse',
    packages=[
        'flask_warehouse',
        'flask_warehouse.backends',
    ],
    package_dir={'flask_warehouse':
                 'flask_warehouse'},
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='flask_warehouse',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import gzip
from flask import Flask

"""
test_flask_warehouse
----------------------------------

Tests for `flask_warehouse` module.
"""

import pytest

from contextlib import contextmanager
from click.testing import CliRunner

from flask_warehouse import Warehouse

from moto import mock_s3


@pytest.fixture
def app():
    app = Flask(__name__)
    return app


@pytest.fixture
def s3_app():
    app = Flask(__name__)

    app.config['WAREHOUSE_DEFAULT_SERVICE'] = 's3'
    app.config['WAREHOUSE_DEFAULT_LOCATION'] = 'us-west-1'

    return app


def test_default_service_is_file(app):
    """Sample pytest test function with the pytest fixture as an argument.
    """
    warehouse = Warehouse(app)

    assert warehouse.service is not None
    assert warehouse.service.id == 'file'

    with app.app_context():
        cubby = warehouse('file:///something/beautiful')
        assert cubby.delete()
        assert not cubby.exists()
        cubby.store(bytes=b'12345')
        assert cubby.exists()
        assert cubby.retrieve() == b'12345'
        assert cubby.delete()


def test_default_service_is_file(app):
    """Sample pytest test function with the pytest fixture as an argument.
    """
    warehouse = Warehouse(app)

    assert warehouse.service is not None
    assert warehouse.service.id == 'file'

    with app.app_context():
        cubby = warehouse('file:///something/beautiful')
        assert cubby.delete()
        assert not cubby.exists()
        cubby.store(bytes=b'12345')
        assert cubby.exists()
        assert cubby.retrieve() == b'12345'
        assert cubby.delete()


@mock_s3
def test_s3_service(app):
    """Sample pytest test function with the pytest fixture as an argument.
    """
    app.config['WAREHOUSE_DEFAULT_SERVICE'] = 's3'
    app.config['WAREHOUSE_DEFAULT_LOCATION'] = 'us-west-1'

    warehouse = Warehouse(app)

    assert warehouse.service is not None
    assert warehouse.service.id == 's3'

    with app.app_context():
        cubby = warehouse('s3:///something/beautiful')
        cubby.bucket.create()

        assert cubby.delete()
        assert not cubby.exists()
        cubby.store(bytes=b'12345')
        assert cubby.exists()
        assert cubby.retrieve() == b'12345'
        assert cubby.filesize() == 5
        assert cubby.mimetype() == None
        cubby.set_mimetype("application/octet-stream")
        assert cubby.mimetype() == 'application/octet-stream'

        assert cubby.metadata() == {}
        metadata = {"tag": "value"}
        assert cubby.set_metadata(metadata)
        assert cubby.metadata() == metadata

        assert cubby.content_encoding() is None
        cubby.set_content_encoding("gzip")
        assert cubby.content_encoding() == 'gzip'

        assert cubby.delete()

@mock_s3
def test_s3_service_gzipped_content(app):
    """Tests gzip support for cubby
    """
    app.config['WAREHOUSE_DEFAULT_SERVICE'] = 's3'
    app.config['WAREHOUSE_DEFAULT_LOCATION'] = 'us-west-1'

    warehouse = Warehouse(app)

    assert warehouse.service is not None
    assert warehouse.service.id == 's3'

    with app.app_context():
        cubby = warehouse('s3:///something/beautiful')
        cubby.bucket.create()

        cubby.store(bytes=b'12345')
        assert cubby.exists()

        # set up attributes for the cubby
        cubby.set_mimetype("application/octet-stream")
        cubby.set_content_encoding("gzip")

        # confirm all attributes are set
        assert cubby.mimetype() == "application/octet-stream"
        assert cubby.content_encoding() == 'gzip'

        ## Assert gzipped content can be read and written correctly
        cubby.store(bytes=b'12345')

        # assert gzipped content was written
        expected_content = gzip.compress(b'12345')
        actual_content = cubby._key.get()['Body'].read()
        # convert byte arrays to lists for comparison
        assert list(expected_content) == list(actual_content)

        # assert no attributes were changed
        assert cubby.mimetype() == 'application/octet-stream'
        assert cubby.content_encoding() == 'gzip'

        # assert can read data correctly
        assert cubby.retrieve() == b'12345'


@mock_s3
def test_s3_service_readme(app):
    """Runs the example from the README."""
    import os

    from flask import Flask
    from flask_warehouse import Warehouse
    
    # 1. Configuring Warehouse
    app = Flask(__name__)
    app.config['WAREHOUSE_DEFAULT_SERVICE'] = 's3'          # or 'file' for filesystem
    app.config['WAREHOUSE_DEFAULT_LOCATION'] = 'us-west-1'  # required for 's3'
    app.config['WAREHOUSE_DEFAULT_BUCKET'] = None

    app.config['AWS_ACCESS_KEY_ID'] = '...'  # required for 's3'
    app.config['AWS_SECRET_ACCESS_KEY'] = '...'  # required for 's3'
    
    warehouse = Warehouse(app)
    
    # Object-oriented approach:
    bucket = warehouse.bucket('mybucket')
    bucket.create()
    oo_cubby = bucket.cubby('keys')

    # Or compact approach:
    compact_cubby = warehouse('s3:///mybucket/keys')

    assert oo_cubby == compact_cubby

    cubby = oo_cubby
    
    # 2. Writing to/from bytes
    contents = b'12345'
    cubby.store(bytes=contents)
    
    assert cubby.filesize() == 5
    
    cubby_contents = cubby.retrieve()
    assert cubby_contents == contents

    # 3. Writing to/from files
    filepath = "local.txt"
    with open(filepath, 'wb') as f:
        f.write(b"Here are the contents of a file.")

    cubby.store(filepath=filepath)
    assert os.path.getsize(filepath) == cubby.filesize()

    assert cubby.retrieve() == open(filepath, 'rb').read()


@mock_s3
def test_copy_move(app):
    """Runs the example from the README."""
    app.config['WAREHOUSE_DEFAULT_SERVICE'] = 's3'
    app.config['WAREHOUSE_DEFAULT_LOCATION'] = 'us-west-1'

    warehouse = Warehouse(app)
    
    # Object-oriented approach:
    source_bucket = warehouse.bucket('source')
    source_bucket.create()

    destination_bucket = warehouse.bucket('destination')
    destination_bucket.create()

    example_source = source_bucket.cubby('example')
    contents = b"Hello, world!"
    example_source.store(bytes=contents)

    example_source.copy_to(key="another")
    another_source = source_bucket.cubby("another")

    assert example_source.retrieve() == another_source.retrieve()

    example_destination = destination_bucket.cubby("destination")

    example_source.copy_to(cubby=example_destination)

    assert example_source.retrieve() == example_destination.retrieve()

    example_source.move_to(cubby=example_destination)
    assert example_destination.retrieve()  == contents
    assert not example_source.exists()


@mock_s3
def test_change_acl(s3_app):
    """Runs the example from the README."""
    warehouse = Warehouse(s3_app)

    source_bucket = warehouse.bucket('source')
    source_bucket.create()

    contents = b"Hello, world!"
    example_source = source_bucket.cubby('example', acl='public-read').store(bytes=contents)
    assert example_source.mimetype() is None

    example_source._key.Acl().load()
    grants = example_source._key.Acl().grants  # grants should be the same before/after

    example_source.set_mimetype("application/json")

    example_source._key.Acl().load()
    assert example_source._key.Acl().grants == grants  # grants should be the same before/after